#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <richedit.h>
#include "AkelDLL.h"


//Functions prototypes
LRESULT CALLBACK NewEditParentProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void CreateEditCaret(HWND hWnd, int nInsertWidth, int nInsertHeight, int nOverwriteWidth, int nOverwriteHeight);
BOOL IsInRect(RECT *rcCanvas, RECT *rcDraw);
void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
HWND hMainWnd;
HWND hMdiClient;
BOOL bOldWindows;
BOOL bOldRichEdit;
BOOL bMDI;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
int nInsertWidth=2;
int nInsertHeight=-1;
int nOverwriteWidth=-1;
int nOverwriteHeight=-1;
POINT ptAkelEditDefault;
WNDPROCDATA *NewMainProcData=NULL;
WNDPROCDATA *NewFrameProcData=NULL;
WNDPROCDATA *NewEditProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(-1, -1, -1, -1);
  pv->pPluginName="Caret";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  //Is plugin already loaded?
  if (bInitMain)
  {
    UninitMain();

    if (pd->hWndEdit)
    {
      SendMessage(pd->hWndEdit, WM_SETFOCUS, (WPARAM)NULL, 0);
    }
  }
  else
  {
    InitMain();

    if (!pd->bOnStart && pd->hWndEdit)
    {
      SendMessage(pd->hWndEdit, WM_SETFOCUS, (WPARAM)NULL, 0);
    }

    //Stay in memory, and show as active
    pd->nUnload=UD_NONUNLOAD_ACTIVE;
  }
}

LRESULT CALLBACK NewEditParentProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_NOTIFY)
  {
    if (wParam == ID_EDIT)
    {
      if (((NMHDR *)lParam)->code == EN_SELCHANGE)
      {
        CreateEditCaret(((NMHDR *)lParam)->hwndFrom, nInsertWidth, nInsertHeight, nOverwriteWidth, nOverwriteHeight);
      }
    }
  }

  //Call next procedure
  if (hWnd == hMainWnd)
    return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
  else
    return NewFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_SETFOCUS ||
      uMsg == WM_PAINT ||
      uMsg == WM_LBUTTONDOWN ||
      (uMsg == WM_KEYDOWN && (wParam == VK_INSERT)))
  {
    LRESULT lResult;

    lResult=NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
    CreateEditCaret(hWnd, nInsertWidth, nInsertHeight, nOverwriteWidth, nOverwriteHeight);
    return lResult;
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

void CreateEditCaret(HWND hWnd, int nInsertWidth, int nInsertHeight, int nOverwriteWidth, int nOverwriteHeight)
{
  CHARRANGE cr={0, -1};
  TEXTMETRICA tm;
  EDITINFO ei;
  RECT rcDraw;
  RECT rcCaret;
  POINT ptCaretChar;
  SIZE sizeAverageWidth;
  HFONT hFont;
  HFONT hFontOld;
  HDC hDC;
  int nWidth=nInsertWidth;
  int nHeight=nInsertHeight;

  if (GetFocus() != hWnd) return;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    if (ei.bOvertypeMode)
    {
      nWidth=nOverwriteWidth;
      nHeight=nOverwriteHeight;
    }
  }
  SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);

  if (cr.cpMin == cr.cpMax)
  {
    if (hDC=GetDC(hWnd))
    {
      if (bOldWindows)
        hFont=CreateFontIndirectA((LOGFONTA *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
      else
        hFont=CreateFontIndirectW((LOGFONTW *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
      hFontOld=(HFONT)SelectObject(hDC, hFont);
      GetTextMetricsA(hDC, &tm);
      if (nWidth == -1)
      {
        GetTextExtentPoint32A(hDC, "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", 52, &sizeAverageWidth);
        nWidth=sizeAverageWidth.cx / 52;
      }
      if (nHeight == -1)
      {
        nHeight=tm.tmHeight - tm.tmExternalLeading;
      }
      if (hFontOld) SelectObject(hDC, hFontOld);
      DeleteObject(hFont);

      SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcDraw);
      rcDraw.right-=rcDraw.left;
      rcDraw.bottom-=rcDraw.top;
      GetCaretPos((POINT *)&rcCaret);
      rcCaret.right=nWidth;
      rcCaret.bottom=nHeight;

      if (IsInRect(&rcDraw, &rcCaret))
      {
        if (rcCaret.left + rcCaret.right > rcDraw.left + rcDraw.right)
          nWidth=(rcDraw.left + rcDraw.right) - rcCaret.left;
        if (rcCaret.top + rcCaret.bottom > rcDraw.top + rcDraw.bottom)
          nHeight=(rcDraw.top + rcDraw.bottom) - rcCaret.top;
        if (rcCaret.top == rcDraw.top)
        {
          GetPosFromChar(hWnd, cr.cpMin, &ptCaretChar);
          nHeight=(ptCaretChar.y + nHeight) - rcDraw.top;
        }
        CreateCaret(hWnd, (HBITMAP)NULL, nWidth, nHeight);
        ShowCaret(hWnd);
      }
      ReleaseDC(hWnd, hDC);
    }
  }
}

BOOL IsInRect(RECT *rcCanvas, RECT *rcDraw)
{
  if (rcDraw->left >= rcCanvas->left + rcCanvas->right ||
      rcDraw->top >= rcCanvas->top + rcCanvas->bottom)
    return FALSE;
  if (rcDraw->left + rcDraw->right <= rcCanvas->left ||
      rcDraw->top + rcDraw->bottom <= rcCanvas->top)
    return FALSE;

  return TRUE;
}

void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt)
{
  DWORD dwPosition;

  if (bOldRichEdit)
  {
    dwPosition=SendMessage(hWnd, EM_POSFROMCHAR, nCharIndex, (LPARAM)NULL);
    pt->x=LOWORD(dwPosition);
    pt->y=HIWORD(dwPosition);
  }
  else
  {
    SendMessage(hWnd, EM_POSFROMCHAR, (WPARAM)pt, nCharIndex);
  }
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hMainWnd=pd->hMainWnd;
  hMdiClient=pd->hMdiClient;
  bOldWindows=pd->bOldWindows;
  bOldRichEdit=pd->bOldRichEdit;
  bMDI=pd->bMDI;
}

void InitMain()
{
  bInitMain=TRUE;

  //SubClass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewEditParentProc, (LPARAM)&NewMainProcData);

  if (bMDI)
  {
    NewFrameProcData=NULL;
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NewEditParentProc, (LPARAM)&NewFrameProcData);
  }

  NewEditProcData=NULL;
  SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NewEditProc, (LPARAM)&NewEditProcData);
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }
  if (bMDI)
  {
    if (NewFrameProcData)
    {
      SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NULL, (LPARAM)&NewFrameProcData);
      NewFrameProcData=NULL;
    }
  }
  if (NewEditProcData)
  {
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NULL, (LPARAM)&NewEditProcData);
    NewEditProcData=NULL;
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
